// Authors:
//  - cadorn, Christoph Dorn <christoph@christophdorn.com>, Copyright 2011, MIT License

var API = require("./api"),
    DESCRIPTORS = require("./descriptors"),
    PACKAGE = require("./package");

var Sandbox = exports.Sandbox = function Sandbox(options)
{
    this.options = options;
    this.loader = null;
    this.packages = {};
}

Sandbox.prototype.clone = function()
{
    var sandbox = new Sandbox(this.options);
    sandbox.packages = this.packages;
    sandbox.setProgram(this.program);
    return sandbox;
}

Sandbox.prototype.setProgram = function(program)
{
    this.program = program;
    
    // Add provider packages
    var packages = this.program.getProviderPackages();
    for (var id in packages)
    {
        this.packages[id] = new PACKAGE.ProviderPackage(id, packages[id]);
    }
    this.init();
}

Sandbox.prototype.init = function()
{
    var self = this;


    // ######################################################################
    // # BravoJS
    // ######################################################################

    var loader = self.loader = {
        mainModuleDir: self.options.mainModuleDir,
        platform: API.ENV.platform,
        api: {
            system: API.SYSTEM,
            errorReporter: function(e)
            {
                API.SYSTEM.print("[BravoJS] " + e + "\n" + e.stack);
            }
        }
    };

    require("./bravojs/bravo").BravoJS(loader);


    // ######################################################################
    // # BravoJS - Plugins
    // ######################################################################

    var Plugin = function() {}

    Plugin.prototype.requireModule = function(id)
    {
        if (!id)
            return;

        // Determine if we are dealing with a provider package
        var pkg = self.packageForId(id, true);
        if (pkg && typeof pkg.isProviderPackage != "undefined" && pkg.isProviderPackage === true)
        {
            return pkg.requireModule(id);
        }

        // If it is a native package we let the platform's require handle it
        if (pkg && pkg.getIsNative() === true)
            return API.ENV.platformRequire(id.replace(/@\//g, "\/"));
    }
    Plugin.prototype.contextForId = function(id)
    {
        if (!id) return;
        try
        {
            var id = self.packageForId(id).path;

            if (typeof this.bravojs.contexts[id] == "undefined")
            {
                this.bravojs.makeContext(id);
            }
            return id;
        }
        catch(e)
        {
            // If this throws the ID was likely a non-packaged module ID
            // We only throw if we should have found a package
            if (id.indexOf("@/") !== -1)
                throw new Error("Unable to find package for ID: " + id);
        }
    }
    Plugin.prototype.loadPackageDescriptor = function(id)
    {
        return self.packageForId(id).normalizedDescriptor.toJSONObject();          
    }
    /**
     * This is called after normalizing the moduleIdentifier and gives the opportunity
     * to throw if an ID was resolved incorrectly due to a missing mapping for example.
     */
    Plugin.prototype.verifyModuleIdentifier = function(moduleIdentifier, args)
    {
        var pkg = self.packageForId(moduleIdentifier, true);

        if (typeof pkg.isProviderPackage != "undefined" && pkg.isProviderPackage === true)
        {
            // ignore for now
        }
        else
        if (typeof pkg.normalizedDescriptor.json.preload != "undefined")
        {
            // ignore for now
            // TODO: Ask preloaders to verify that the source exists
        }
        else
        {
            if (!pkg.moduleSourceExists(moduleIdentifier))
            {
                throw new Error("Incorrectly resolved (file does not exist) moduleIdentifier '" + API.UTIL.locatorToString(args.moduleIdentifier) + "' to '" + API.UTIL.locatorToString(moduleIdentifier) + "' against context '" + args.context.id + "' and relativeModuleDir '" + args.relativeModuleDir + "'.");
            }
        }
    }
    loader.bravojs.registerPlugin(new Plugin());
    loader.bravojs.registerPlugin(new (require("./bravojs/plugins/packages/packages").Plugin)());


    // ######################################################################
    // # BravoJS - Module constructor prototypes
    // ######################################################################

    var loading;

    loader.bravojs.module.constructor.prototype.load = function pinf_loader_load(moduleIdentifier, callback)
    {
        if (typeof moduleIdentifier == "object")
        {
            if (API.DEBUG.enabled)
                if (API.ENV.booting)
                    API.SYSTEM.print("\0)\0)\0magenta(\0:blue(<=====\0:)\0)\n");
                else
                    API.SYSTEM.print("\0magenta(\n");

            var t = new Date().getTime();

            // Load an extra package into the program including all its dependencies
            // and start with the main module
            self.program.assembler.addPackageToProgram(self, self.program, moduleIdentifier, function(pkg)
            {
                loader.bravojs.module.constructor.prototype.load(pkg.getMainId(moduleIdentifier), function(moduleIdentifier)
                {
                    if (API.DEBUG.enabled)
                        if (API.ENV.booting)
                            API.SYSTEM.print("\0magenta(\0:blue(=====>\0:)\0)\0green(\0bold(", false, true);
                        else
                            API.SYSTEM.print("\0)");

                    API.ENV.timers.loadAdditional += new Date().getTime() - t;

                    callback(moduleIdentifier);
                });
            })
            return;
        }
        else
        if (moduleIdentifier.charAt(0)==".")
        {
            throw new Error("Relative IDs '" + moduleIdentifier + "' to module.load() not supported at this time.");
        }

        var context = loader.bravojs.contextForId(moduleIdentifier, true);
        moduleIdentifier = context.resolveId(moduleIdentifier);

        function load(data)
        {
            loading = {
                id: moduleIdentifier,
                callback: function()
                {
                    callback(moduleIdentifier);
                }
            };

            try
            {
                // Remove shebang line if present
                data = data.replace(/^#!(.*?)\n/, "");

                var m;
                if (/^text!/.test(moduleIdentifier))
                {
                    // text plugin
                    eval(
                        "loader.bravojs.module.declare([], function() {\n" +
                        'return ["' + data.replace(/"/g, '\\"').replace(/\n/g, '","') + '"].join("\\n");' +
                        "\n});"
                    );
                }
                else
                if (m = data.match(/(^|[\r\n])\s*define\(\[([^\]]*)\].*/))
                {
                    // AMD with dependencies

                    var deps = m[2].split(",").map(function(dep)
                    {
                        return "'" + dep.replace(/^\s*"|^\s*'|"\s*$|'\s*$/g, "") + "'";
                    });

                    // strip all data prior to define()
                    var parts = data.split(m[0]);
                    if (parts.length != 2)
                        throw new Error("Unable to fix define() wrapper for module: " + moduleIdentifier);
                    data = m[0].replace(/^[\r\n]\s*/, "") + parts[1];

                    eval(
                        "loader.bravojs.module.declare([" +
                        deps.filter(function(dep)
                        {
                            return (!/^'(require|module|exports)'$/.test(dep));
                        }).join(",") +
                        "], function(require, exports, module) {\n" + 
                        "({define:function(deps, factory){module.exports = factory(" +
                        deps.map(function(dep)
                        {
                            if (/^'(require|module|exports)'$/.test(dep))
                                return dep.substring(1, dep.length-1);
                            else
                                return "require(" + dep + ")";
                        }).join(",") +
                        ");}})." +
                        data +
                        "\n});"
                    );
                }
                else
                if (m = data.match(/(^|[\r\n])\s*define\(\s*function\s*\(\s*require\s*,\s*exports\s*,\s*module\s*\)\s*{.*/))
                {
                    // AMD without dependencies

                    // strip all data prior to define()
                    var parts = data.split(m[0]);
                    if (parts.length != 2)
                        throw new Error("Unable to fix define() wrapper for module: " + moduleIdentifier);
                    data = parts[1];

                    eval(
                        "loader.bravojs.module.declare([" +
                        API.UTIL.scrapeDeps(data).join(',') +
                        "], function(require, exports, module) {\n" + 
                        data
                    );
                }
                else
                if ((typeof loader.bravojs.module.constructor.prototype.load != "undefined" &&
                     typeof loader.bravojs.module.constructor.prototype.load.modules11 != "undefined" &&
                     loader.bravojs.module.constructor.prototype.load.modules11 === false) || data.match(/(^|[\r\n])\s*module.declare\s*\(/))
                {
                    // Modules/2
                    eval("loader.bravojs." + data.replace(/^\s\s*/g, ""));
                }
                else
                {
                    // Modules/1.1
                    eval("loader.bravojs.module.declare([" + API.UTIL.scrapeDeps(data).join(',') + "], function(require, exports, module) {\n" + data + "\n});");
                }
            }
            catch(e)
            {
                e.message += "\n in module " + moduleIdentifier;
                throw e;
            }
        }

        var pkg = self.packageForId(moduleIdentifier, true);
        if (pkg)
        {
            // We do not need to load native modules. See Plugin.prototype.requireModule above.
            if (pkg.getIsNative())
            {
                callback(moduleIdentifier);
            }
            else
            if (pkg.getModuleSource(self, moduleIdentifier, load) === false)
            {
                callback(moduleIdentifier);
            }
        }
        else
        {
            var URL = loader.bravojs.require.canonicalize(moduleIdentifier),
                m = URL.match(/^memory:\/(.*)$/),
                path = m[1];

            path = path.replace(/^\w*!/, "");

            if (/\.js$/.test(path) && !API.FILE.exists(path))
                path = path.substring(0, path.length-3);

            load(API.FILE.read(path));
        }
    }

    loader.bravojs.module.constructor.prototype.declare = function pinf_loader_declare(dependencies, moduleFactory)
    {
        var id    = loading.id;
        var callback  = loading.callback;

        loading = void 0;

        if (typeof dependencies === "function")
        {
          moduleFactory = dependencies;
          dependencies = [];
        }

        loader.bravojs.provideModule(dependencies, moduleFactory, id, callback);
    }


    // ######################################################################
    // # Sandbox API
    // ######################################################################
    
    self.declare = loader.bravojs.module.declare;
}

/**
 * Create or get existing package for path
 * 
 * NOTE: The 'locator' argument gets modified!
 * TODO: Refactor so 'locator' argument does not get modified.
 */
Sandbox.prototype.ensurePackageForLocator = function(locator, options)
{
    var self = this;
    var locatorDescriptor = locator.descriptor || void 0;
    delete locator.descriptor;

    function finalize(packageId, newLocator)
    {
        if (typeof newLocator == "undefined")
        {
            if (typeof self.packages[packageId].uid != "undefined")
                locator.uid = self.packages[packageId].uid;
            newLocator = self.program.descriptor.augmentLocator(locator, options);
        }
        for (var key in newLocator)
            locator[key] = newLocator[key];
        if (typeof locatorDescriptor != "undefined")
            locator.descriptor = locatorDescriptor;
        return self.packages[packageId];
    }

    if (typeof locator.id != "undefined" && typeof this.packages[locator.id] != "undefined")
    {
        return finalize(locator.id);
    }
    else
    if (typeof locator.uid != "undefined" && typeof this.packages[locator.uid] != "undefined")
    {
        return finalize(locator.uid);
    }
    var path = locator.location;
    if (typeof this.packages[path] == "undefined")
    {
        this.packages[path] = new PACKAGE.Package(new DESCRIPTORS.Package(path));

        // If package has a UID set we also index our packages by it
        // TODO: Add version to key if applicable
        if (typeof this.packages[path].uid != "undefined")
        {
            locator.id = this.packages[path].uid;
            this.packages[this.packages[path].uid] = this.packages[path];
        }

        // If locator has an ID set we also index our packages by it
        if (typeof locator.id != "undefined")
            this.packages[locator.id] = this.packages[path];

        // Merge descriptor information from the locator onto the package descriptor if applicable
        // We first ask the program descriptor to augment the locator with any additional info
        var newLocator = this.program.descriptor.augmentLocator(locator, options);

        if (typeof newLocator.descriptor != "undefined")
        {
            API.UTIL.deepMerge(this.packages[path].normalizedDescriptor.json, newLocator.descriptor);
        }

        // Convert mapped module IDs to paths
        if (typeof this.packages[path].normalizedDescriptor.json.modules != "undefined")
        {
            var libDir = this.packages[path].normalizedDescriptor.json.directories && this.packages[path].normalizedDescriptor.json.directories.lib;
            if (typeof this.libDir != "string")
                libDir = "lib";

            var modules = {};
            for (var id in this.packages[path].normalizedDescriptor.json.modules)
            {
                if (id.charAt(0) == "." || id.charAt(0) == "/")
                    modules[id.replace(/^\./, "")] = this.program.descriptor.augmentLocator(this.packages[path].normalizedDescriptor.json.modules[id], options);
                else
                    modules["/" + libDir + "/" + id] = this.program.descriptor.augmentLocator(this.packages[path].normalizedDescriptor.json.modules[id], options);
            }
            this.packages[path].normalizedDescriptor.json.modules = modules;
        }
        return finalize(path, newLocator);
    }
    else
        return finalize(path);
}

/**
 * Get an existing package for id
 */
Sandbox.prototype.packageForId = function(id, silent)
{
    if (!id)
        throw new Error("Empty ID!");

    var m = id.match(/^(\/?)(.*?)(@\/(.*))?$/),
        lookupIds;

    // m[1] - '/' prefix
    // m[2] - path
    // m[3] -
    // m[4] - after @/

    if (!m[1] && m[2])
        lookupIds = [ m[2] + "/", m[2] ];
    else
    if (m[1] == "/" && m[2])
        lookupIds = [ "/" + m[2] + "/", m[2] + "/", "/" + m[2], m[2]];

    var lookupId;
    if (!lookupIds || lookupIds.length==0 || (lookupId = Object.keys(this.packages).filter(function(id) { return (lookupIds.indexOf(id)>-1); })).length == 0)
    {
        if (silent)
            return null;
        throw new Error("Package for id '" + id + "' not found via lookup IDs '" + lookupIds + "' in packages: " + Object.keys(this.packages));
    }
    return this.packages[lookupId[0]];
}
